/* 
 * Copyright (C) 2006 Darren Salt <linux@youmustbejoking.demon.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
 * Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
                     
#include <gtk/gtkicontheme.h>
#include <gtk/gtkmenu.h>
#include <gtk/gtkimagemenuitem.h>

#include "gtk_icon_theme_menu.h"
#include "gtk_icon_theme_context_menu.h"

// FIXME
#define _(x) (x)

static void gtk_icon_theme_menu_class_init (GtkIconThemeMenuClass *);
static void gtk_icon_theme_menu_init (GtkIconThemeMenu *);
static void gtk_icon_theme_menu_set_property (GObject *, guint, const GValue *,
					      GParamSpec *);
static void gtk_icon_theme_menu_get_property (GObject *, guint, GValue *,
					      GParamSpec *);
static void gtk_icon_theme_menu_destroy (GtkObject *);

static void gtk_icon_theme_menu_update (GtkIconThemeMenu *);

#define GET_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE ((o), gtk_icon_theme_menu_get_type (), \
				GtkIconThemeMenuPrivate))

enum
{
  PROP_0,
  PROP_THEME,
  PROP_THEME_NAME,
  PROP_CONTEXTS,
  PROP_LAST
};

typedef struct _GtkIconThemeMenuPrivate GtkIconThemeMenuPrivate;
struct _GtkIconThemeMenuPrivate
{
  gboolean destroyed;
  GtkIconTheme *theme;
  gchar *contexts;
};

static GtkMenuClass *parent_class = NULL;

enum
{
  SIGNAL_ACTIVATED,
  SIGNAL_LAST
};

static guint signals[SIGNAL_LAST];

static const char default_contexts[] =
  "Actions "
  "Applications "
  "Categories "
  "Devices "
  "Emblems "
  "Emotes "
  "FileSystems "
  "MimeTypes "
  "Misc "
  "Places "
  "Status "
  "Stock "
  ;

GType
gtk_icon_theme_menu_get_type (void)
{
  static GType type = 0;

  if (!type)
  {
    static const GTypeInfo info = {
      sizeof (GtkIconThemeMenuClass),
      NULL,			/* base_init */
      NULL,			/* base_finalize */
      (GClassInitFunc) gtk_icon_theme_menu_class_init,
      NULL,			/* class_finalize */
      NULL,			/* class_data */
      sizeof (GtkIconThemeMenu),
      0,			/* n_preallocs */
      (GInstanceInitFunc) gtk_icon_theme_menu_init,
      NULL			/* value_table */
    };

    type = g_type_register_static (GTK_TYPE_MENU, "GtkIconThemeMenu", &info, 0);
  }

  return type;
}

static void
gtk_icon_theme_menu_class_init (GtkIconThemeMenuClass *klass)
{
  GObjectClass *gobject_class;
  GtkObjectClass *object_class;
  GtkIconThemeMenuClass *itm_class;

  gobject_class = (GObjectClass *) klass;
  object_class = (GtkObjectClass *) klass;
  itm_class = (GtkIconThemeMenuClass *) klass;

  parent_class = g_type_class_peek_parent (klass);

  object_class->destroy = gtk_icon_theme_menu_destroy;

  itm_class->activate = NULL;

  gobject_class->set_property = gtk_icon_theme_menu_set_property;
  gobject_class->get_property = gtk_icon_theme_menu_get_property;

  signals[SIGNAL_ACTIVATED] =
    g_signal_new ("activate",
	G_TYPE_FROM_CLASS (object_class),
	G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
	G_STRUCT_OFFSET (GtkIconThemeMenuClass, activate),
	NULL, NULL,
	g_cclosure_marshal_VOID__STRING,
	G_TYPE_NONE, 1, G_TYPE_STRING);

  g_object_class_install_property
    (gobject_class, PROP_THEME,
     g_param_spec_object ("theme", _("Theme"),
			  _("The icon theme listed in the menu"),
			  gtk_icon_theme_menu_get_type (),
			  G_PARAM_READABLE | G_PARAM_WRITABLE));

  g_object_class_install_property
    (gobject_class, PROP_THEME_NAME,
     g_param_spec_string ("theme-name", _("Theme name"),
			  _("The name of the icon theme listed in the menu"),
			  NULL, G_PARAM_WRITABLE));

  g_object_class_install_property
    (gobject_class, PROP_CONTEXTS,
     g_param_spec_string ("contexts", _("Contexts"),
			  _("A space-separated list of context names for the submenus (NULL for the default)"),
			  NULL, G_PARAM_READABLE | G_PARAM_WRITABLE));

  g_type_class_add_private (gobject_class, sizeof (GtkIconThemeMenuPrivate));  
}

static void
gtk_icon_theme_menu_init (GtkIconThemeMenu *menu)
{
  GtkIconThemeMenuPrivate *priv = GET_PRIVATE (menu);
  priv->destroyed = FALSE;
  priv->theme = NULL;
  priv->contexts = g_strdup (default_contexts);
}

static void
gtk_icon_theme_menu_set_property (GObject *object, guint prop_id,
				  const GValue *value, GParamSpec *pspec)
{
  g_return_if_fail (GTK_IS_ICON_THEME_MENU (object));
  GtkIconThemeMenu *menu = (GtkIconThemeMenu *) object;

  switch (prop_id)
  {
  case PROP_THEME:
    gtk_icon_theme_menu_set_theme (menu, g_value_get_object (value));
    break;

  case PROP_THEME_NAME:
    gtk_icon_theme_menu_set_named_theme (menu, g_value_get_string (value));
    break;

  case PROP_CONTEXTS:
    {
      GtkIconThemeMenuPrivate *priv = GET_PRIVATE (object);
      g_free (priv->contexts);
      priv->contexts = g_strdup (g_value_get_string (value));
      gtk_icon_theme_menu_update (menu);
    }
    break;

  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

static void
gtk_icon_theme_menu_get_property (GObject *object, guint prop_id,
				  GValue *value, GParamSpec *pspec)
{
  GtkIconThemeMenu *menu = GTK_ICON_THEME_MENU (object);
  GtkIconThemeMenuPrivate *priv = GET_PRIVATE (menu);

  switch (priv ? prop_id : 0)
  {
  case PROP_THEME:
    g_value_set_object (value, priv->theme);
    break;

  case PROP_CONTEXTS:
    g_value_set_string (value, priv->contexts);
    break;

  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

static void
gtk_icon_theme_menu_destroy (GtkObject *object)
{
  GtkIconThemeMenu *menu = GTK_ICON_THEME_MENU (object);
  g_return_if_fail (menu != NULL);

  GtkIconThemeMenuPrivate *priv = GET_PRIVATE (menu);

  if (priv->destroyed)
    return;

  priv->destroyed = TRUE;

  if (priv->theme)
    g_object_unref (G_OBJECT (priv->theme));
  g_free (priv->contexts);

  (*GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}

/**
 * gtk_icon_theme_menu_new:
 * @returns: A new #GtkIconThemeMenu.
 *
 * Creates a new #GtkIconThemeMenu with the default theme.
 **/
GtkWidget *
gtk_icon_theme_menu_new (void)
{
  GtkIconThemeMenu *menu = g_object_new (gtk_icon_theme_menu_get_type (), NULL);
  gtk_icon_theme_menu_set_theme (menu, gtk_icon_theme_get_default ());
  return GTK_WIDGET (menu);
}

/**
 * gtk_icon_theme_menu_new_with_theme:
 * @theme: A #GtkIconTheme.
 * @returns: A new #GtkIconThemeMenu.
 *
 * Creates a new #GtkIconThemeMenu with a custom theme.
 **/
GtkWidget *
gtk_icon_theme_menu_new_with_theme (GtkIconTheme *theme)
{
  GtkIconThemeMenu *menu = g_object_new (gtk_icon_theme_menu_get_type (), NULL);
  gtk_icon_theme_menu_set_theme (menu, theme);
  return GTK_WIDGET (menu);
}

/**
 * gtk_icon_theme_menu_new_with_named_theme:
 * @theme: The name of an icon theme.
 * @returns: A new #GtkIconThemeMenu.
 *
 * Creates a new #GtkIconThemeMenu with a custom theme.
 **/
GtkWidget *
gtk_icon_theme_menu_new_with_named_theme (const gchar *theme)
{
  GtkIconThemeMenu *menu = g_object_new (gtk_icon_theme_menu_get_type (), NULL);
  GtkIconThemeMenuPrivate *priv = GET_PRIVATE (menu);
  priv->theme = gtk_icon_theme_new ();
  g_object_ref (G_OBJECT (priv->theme));
  gtk_icon_theme_menu_set_named_theme ((GtkIconThemeMenu *) menu, theme);
  return GTK_WIDGET (menu);
}

/**
 * gtk_icon_theme_menu_set_theme:
 * @menu: A #GtkIconThemeMenu, created using #gtk_icon_theme_menu_new_with_theme().
 * @theme: A #GtkIconTheme.
 *
 * Sets @theme as the theme whose icons are listed in the menu.
 **/
void
gtk_icon_theme_menu_set_theme (GtkIconThemeMenu *menu, GtkIconTheme *theme)
{
  g_return_if_fail (GTK_IS_ICON_THEME_MENU (menu));
  if (theme)
  {
    g_return_if_fail (GTK_IS_ICON_THEME (theme));
  }
  GtkIconThemeMenuPrivate *priv = GET_PRIVATE (menu);
  if (priv->theme)
    g_object_unref (G_OBJECT (priv->theme));
  priv->theme = theme;
  if (theme)
    g_object_ref (G_OBJECT (priv->theme));
  gtk_icon_theme_menu_update (menu);
}

/**
 * gtk_icon_theme_menu_set_named_theme:
 * @menu: A #GtkIconThemeMenu, created using #gtk_icon_theme_menu_new_with_theme().
 * @theme: The name of the icon theme.
 *
 * Sets @theme as the theme whose icons are listed in the menu.
 * Will fail if the menu is using a default icon theme object.
 **/
void
gtk_icon_theme_menu_set_named_theme (GtkIconThemeMenu *menu, const gchar *theme)
{
  g_return_if_fail (GTK_IS_ICON_THEME_MENU (menu));
  GtkIconThemeMenuPrivate *priv = GET_PRIVATE (menu);
  gtk_icon_theme_set_custom_theme (priv->theme, theme);
  gtk_icon_theme_menu_update (menu);
}

static int
qsort_strcmp (const void *a, const void *b)
{
  char *s1 = *(char *const *)a;
  char *s2 = *(char *const *)b;
  return !*s1 ? 1 : !*s2 ? -1 : strcmp (s1, s2);
}

/**
 * gtk_icon_theme_menu_set_contexts:
 * @menu: A #GtkIconThemeMenu, created using #gtk_icon_theme_menu_new_with_theme().
 * @contexts: The icon theme contexts (NULL-terminated array of strings), determining which icons are listed.
 *
 * Sets @contexts as the context list which controls which of the theme's
 * icons are listed in the submenus.
 **/
void
gtk_icon_theme_menu_set_contexts (GtkIconThemeMenu *menu, gchar **contexts)
{
  g_return_if_fail (GTK_IS_ICON_THEME_MENU (menu));

  GtkIconThemeMenuPrivate *priv = GET_PRIVATE (menu);
  g_free (priv->contexts);
  guint length = g_strv_length (contexts);
  gchar **new = g_new0 (gchar *, length + 1);
  for (guint i = 0; i < length; ++i)
    new[i] = contexts[i];

  if (length > 1)
  {
    guint i, j, k;
    qsort (new, length, sizeof (gchar *), qsort_strcmp);

    for (i = 0, j = k = 1; j < length; ++j)
      if (*new[j] && strcmp (new[i], new[j]))
      {
	new[i = k] = new[j];
	++k;
      }
    new[k] = NULL;
  }

  priv->contexts = g_strjoinv (" ", (length && *new[0]) ? new : NULL);
  g_free (new);
  gtk_icon_theme_menu_update (menu);
}

/**
 * gtk_icon_theme_menu_add_contexts:
 * @menu: A #GtkIconThemeMenu, created using #gtk_icon_theme_menu_new_with_theme().
 * @contexts: The extra icon theme contexts (NULL-terminated array of strings), determining which icons are listed.
 *
 * Adds @contexts to the context list which controls which of the theme's
 * icons are listed in the submenus.
 **/
void
gtk_icon_theme_menu_add_contexts (GtkIconThemeMenu *menu, gchar **contexts)
{
  g_return_if_fail (GTK_IS_ICON_THEME_MENU (menu));
  if (!contexts || !*contexts)
    return;

  GtkIconThemeMenuPrivate *priv = GET_PRIVATE (menu);
  gchar **old = g_strsplit (priv->contexts, " ", 0);
  guint old_length = g_strv_length (old);
  guint new_length = g_strv_length (contexts);
  gchar **new = g_new0 (gchar *, old_length + new_length + 1);

  /* shallow-copy the strings */
  guint i;
  for (i = 0; i < old_length; ++i)
    new[i] = old[i];
  for (i = 0; i < new_length; ++i)
    new[i + old_length] = contexts[i];

  gtk_icon_theme_menu_set_contexts (menu, new);
  g_strfreev (old);
  g_free (new);
}

/**
 * gtk_icon_theme_menu_remove_contexts:
 * @menu: A #GtkIconThemeMenu, created using #gtk_icon_theme_menu_new_with_theme().
 * @contexts: The extra icon theme contexts (NULL-terminated array of strings), determining which icons are listed.
 *
 * Adds @contexts to the context list which controls which of the theme's
 * icons are listed in the submenus.
 **/
void
gtk_icon_theme_menu_remove_contexts (GtkIconThemeMenu *menu, gchar **contexts)
{
  g_return_if_fail (GTK_IS_ICON_THEME_MENU (menu));
  if (!contexts || !*contexts)
    return;

  GtkIconThemeMenuPrivate *priv = GET_PRIVATE (menu);
  gchar **strv = g_strsplit (priv->contexts, " ", 0);
  guint i, j;

  for (i = j = 0; strv[i]; ++i)
  {
    guint k;
    for (k = 0; contexts[k]; ++k)
      if (!strcmp (strv[i], contexts[k]))
        break;
    if (contexts[k])
      g_free (strv[i]);
    else
      strv[j++] = strv[i];
  }
  strv[j] = NULL;

  gtk_icon_theme_menu_set_contexts (menu, strv);
  g_strfreev (strv);
}

/**
 * gtk_icon_theme_menu_get_theme:
 * @app_menu_item: An #GtkIconThemeMenu.
 * @returns: A #GtkIconTheme.
 *
 * Returns the current theme set for the #GtkIconThemeMenu. It is referenced
 * by the menu.
 **/
GtkIconTheme *
gtk_icon_theme_menu_get_theme (GtkIconThemeMenu *menu)
{
  g_return_val_if_fail (GTK_IS_ICON_THEME_MENU (menu), NULL);

  GtkIconThemeMenuPrivate *priv = GET_PRIVATE (menu);
  return priv->theme;
}

/**
 * gtk_icon_theme_menu_get_contexts:
 * @app_menu_item: An #GtkIconThemeMenu.
 * @returns: A array of context name strings.
 *
 * Returns the current context set for the #GtkIconThemeMenu. Free with
 * g_strfreev().
 **/
gchar **
gtk_icon_theme_menu_get_contexts (GtkIconThemeMenu *menu)
{
  g_return_val_if_fail (GTK_IS_ICON_THEME_MENU (menu), NULL);

  GtkIconThemeMenuPrivate *priv = GET_PRIVATE (menu);
  return g_strsplit (priv->contexts, " ", 0);
}

static void
gtk_icon_theme_menu_activate (GtkMenuItem *item G_GNUC_UNUSED,
			      const gchar *icon, gpointer data)
{
  g_signal_emit (G_OBJECT (data), signals[SIGNAL_ACTIVATED], 0, icon);
}

static void
gtk_icon_theme_menu_update (GtkIconThemeMenu *menu)
{
  GList *item, *items = gtk_container_get_children (GTK_CONTAINER (menu));

  /* flush the menu */
  for (item = items; item; item = g_list_next (item))
    gtk_container_remove (GTK_CONTAINER (menu), item->data);
  g_list_free (items);

  GtkIconThemeMenuPrivate *priv = GET_PRIVATE (menu);

  gtk_widget_push_composite_child ();

  gchar **contexts = g_strsplit (priv->theme ? priv->contexts : "", " ", 0);
  int i;
  for (i = 0; contexts[i]; ++i)
  {
    if (!*contexts[i])
      continue;

    GtkWidget *icn = gtk_menu_item_new_with_label (contexts[i]);
    gtk_widget_show (icn);
    gtk_widget_set_composite_name (icn, "named_icon_menu_item_menu");
    gtk_menu_shell_append (GTK_MENU_SHELL (menu), icn);

    GtkWidget *sub = gtk_icon_theme_context_menu_new_with_theme (priv->theme);
    gtk_icon_theme_context_menu_set_context (GTK_ICON_THEME_CONTEXT_MENU (sub),
					     contexts[i]);
    gtk_widget_show (sub);
    gtk_menu_item_set_submenu (GTK_MENU_ITEM (icn), sub);

    g_signal_connect (G_OBJECT (sub), "activate",
		      G_CALLBACK (gtk_icon_theme_menu_activate), menu);
  }
  if (!i)
  {
    GtkWidget *icn = gtk_menu_item_new_with_label (_("(empty)"));
    gtk_widget_show (icn);
    gtk_widget_set_composite_name (icn, "named_icon_menu_item_menu");
    gtk_menu_shell_append (GTK_MENU_SHELL (menu), icn);
  }

  gtk_widget_pop_composite_child ();

  g_strfreev (contexts);
}
